;; Group.scm
;;
;; This file contains several groupware components.
;;
;;  a shared textarea -- which allows one to allow others to view your textarea
;;  a remote textarea -- which allows one to view someones shared textarea
;;  a group textarea -- which allows two users to simultaneously edit a document
;;
;; Use:
;;   (define S (make-shared-textarea GROUP-CLIENT NAME-FOR-TEXTAREA ROWS COLS ...))
;;   (define R (make-remote-textarea GROUP-CLIENT NAME-FOR-TEXTAREA ROWS COLS ...))
;;   (define B (make-both-textarea GROUP-CLIENT NAME-FOR-TEXTAREA ROWS COLS ...))
;;
;; Methods:
;;   (S 'get 'textarea) --> returns the underlying textarea
;;   (S 'get 'group-client) --> returns the underlying group-client
;;   (S 'get 'name) --> returns the underlying name
;; We will add methods to connect/disconnect etc.



(jlib.Swing.load)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; (shared-textarea GROUP TEXTAREANAME ROWS COLS . ARGS)
;;   creating and maintaining a shared textarea. 
;;   This will create a textarea where all insert/delete events
;;   are broadcast to everyone in the GROUP
;;   Note that GROUP is a group-client object
;;
;;  The other members of the group should be using a
;;  remote-textarea object which will display all insert/delete
;;  events.
;;
;;  The textarea name is needed because several shared-textareas
;;  may share the same GROUP network object.
;;
;;  A shared-textarea is an object S with several methods you can call
;;   (S 'get 'textarea) -- returns the textarea object itself
;;   (S 'get 'name) -- returns the name
;;   (S 'get 'group-client) -- returns the group-client G
;;  Later we may add methods to allow the textarea to
;;  reconnect to another server, or to disconnect from one
;;  or more remote-textareas
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;; this creates a shared-textarea object and returns just the
;; textarea part
(define shared-textarea (lambda (G name rows cols . args)
  (let ((S
         (apply make-shared-textarea (append (list G name rows cols)  args))
       ))
    (S 'get 'textarea))))


;; this creates the shared textarea object
;; which consists of the textarea and its network state as well
(define make-shared-textarea (lambda (G name rows cols . args)
  (letrec (
         (insertaction (lambda(offset str attr)
;             (display (list "sending insert" offset str G name))(newline)
             (G 'send (list name 'insert 'shared-textarea)  offset (.toString str)) 
             (highlight offset) #t))
         (deleteaction (lambda(offset len str)
;             (display (list "sending delete" offset len str G name))(newline)
             (G 'send (list name 'delete 'shared-textarea)  offset len (.toString str)) 
             (highlight offset) #t))
         (shared-ta
           (apply swingcomp (cons (javax.swing.JTextArea. 
               (grouptool.tatool.PlainDocumentWrapper. insertaction deleteaction)) args)))
         (highlight (make-highlighter shared-ta)))
       
   (.setRows shared-ta rows)
   (.setColumns shared-ta cols)
   (G 'add-listener (list name 'get-contents 'shared-textarea) (lambda(cmd) 
;       (display (list 'sending 'contents 'name 'shared-textarea cmd (.getText shared-ta))) (newline)
       (java.lang.Thread.sleep 100L)
       (G 'send (list name 'contents 'shared-textarea) (.getText shared-ta))))
;   (display (list "adding listener for sharedtextarea" name ))(newline)

    (lambda (cmd . args)
       (case cmd
         ((get) 
           (case (first args)
             ((textarea) shared-ta)
             ((name) name)
             ((group-client) G)
             (else (list 'error "unknown args for get: " args))))
         (else
           (list 'error "unknown command:" cmd args))))
  )))



(define remote-textarea (lambda (G name . args)

   (let* (
          (initialized #f)
          (ta (apply textarea args))
          (highlight (make-highlighter ta)))
     (.addKeyListener ta silencer)

;     (display (list 'add-listener G (list name 'insert 'shared-textarea))) (newline)

     (G 'add-listener (list name 'insert 'shared-textarea) (lambda(cmd offset str)
;        (display (list 'zzz name 'insert 'shared-textarea cmd offset str)) (newline)
        (if (not initialized)
            (begin
;               (display (list "sending get-contents/ins" offset str G name))(newline)
               (G 'send (list name 'get-contents 'shared-textarea)))
            (begin
              (.insertString (.getDocument ta) offset str #null)
              (highlight offset)
        ))))
     (G 'add-listener (list name 'delete 'shared-textarea) (lambda(cmd offset len str)
;        (display (list 'yyy name 'delete 'shared-textarea cmd offset str)) (newline)
        (if (not initialized)
            (begin
;                (display (list "sending get-contents/del" offset len str G name))(newline)
                (G 'send (list name 'get-contents 'shared-textarea)))
            (begin
                (.remove (.getDocument ta) offset len)
                (highlight offset))
            )))
     (G 'add-listener (list name 'contents 'shared-textarea) (lambda(cmd x)
;        (display (list 'xxx name 'contents 'shared-textarea cmd x)) (newline)
         (.setText ta x)
         (set! initialized #t)))

    ta)))





;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; create higlighter for both-textarea
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;





;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;     (both-textarea GROUP-CLIENT NAME COLOR ROWS COLS ARGS)
;;;;
;;;;
;;;; this creates a textarea that is shared by any number
;;;; of users, each using a different color
;;;; The "green" user is the owner and sends logging data
;;;; to the network.
;;;; Currently the highlighting will only work correctly 
;;;; when there are two users: green and blue, but this
;;;; should not be hard to generalize. The only tricky part
;;;; is changing colors when two people are working close
;;;; to each other.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define both-textarea (lambda (G name color rows cols . args)
  (letrec (
     (history (java.util.Vector. 1024))
     (historysize 0)
     (insertaction (lambda(offset str attr)
	(set! string (.toString str))
	(synchronize 'typeahead
		     (lambda(_)
	(G 'send (list name 'insert 'both-textarea)  offset color string historysize)))
	#f))
     (deleteaction (lambda(offset len str)
	(synchronize 'typeahead
	  (lambda(_)
	    (G 'send (list name 'delete 'both-textarea) offset color len (.toString str) historysize)))
	#f))

      ;; this procedure modifies a given insertion/deletion to take into account
      ;; the actions that have been performed while it was in transit
     (compute-true-operation (lambda( action histpoint offset len)
       (if (equal? action 'insert)

         (let loop ((h histpoint) (s historysize) (p offset))
;           (display (list "in ins loop" h s p)) (newline)
           (if (>= h s) 
               (list p len)
               (let* ((hrecord (.get history h))
                      (haction (first hrecord))
                      (hoffset (second hrecord))
                      (hsize (third hrecord)))
                 (if (> hoffset p) ;; ignore operations that appear after your insert point
                     (loop (+ h 1) s p)
                     (case haction
                       ((insert)  ;; handle an insertion before where you wanted to insert
                         (loop (+ h 1) s (+ p hsize)))
                       ((delete)  ;; handle a deletion before where you wanted to insert
                         (cond
                            ((<= (+ hoffset hsize) p)
                             (loop (+ h 1) s (- p hsize)))
                            (else
                             (loop (+ h 1) s hoffset)))))))))

         (let loop ((h histpoint) (s historysize) (p offset) (n len))
;          (display (list "in del loop" h s p n)) (newline)
          (if (<= len 0) (list p len) ;; zero or negative length deletions are ignored
           (if (>= h s) (list p n)
               (let* ((hrecord (.get history h))
                      (haction (first hrecord))
                      (hoffset (second hrecord))
                      (hsize (third hrecord))
                      (hmax (+ hoffset hsize))
                      (pmax (+ p n)))
;                 (display (list "deleting" hrecord "h=" hoffset hmax "p=" p pmax))(newline)
                 (cond ((> hoffset pmax)          ;; ignore operations past your deletion region
                        (loop (+ h 1) s p n))
                       ((equal? haction 'insert)  ;; handle insertions before your deletion region
                        (loop (+ h 1) s (+ p hsize) n))
                       ((< hmax p)                ;; handle deletions that appear entirely before your deletion
                        (loop (+ h 1) s (- p hsize) n))
                       (else  
                                                  ;; handle deletions whose region inserts your deletion region
                        (let* ((below-intersect (max 0 (- hoffset p)))
                               (above-intersect (max 0 (- pmax hmax)))
                               (size-of-intersection (+ below-intersect above-intersect)))
;                           (display (list "in del case" below-intersect above-intersect size-of-intersection h s p n)) (newline)
                           (loop (+ h 1) s hoffset size-of-intersection))))))))
     )))

            
   (do-op (lambda (action offset str len hist)
;        (display (list 'calling-do-op action offset str len (equal? action 'insert)))(newline)
        (if (equal? action 'insert)
           (begin
;              (display "about to insert")(newline)
	      (.insertStringNoSend (.getDocument both-ta) offset str #null 
                 (lambda R
;                    (display (list 'inside-do-op action offset str len))(newline)
                    (set! historysize (hist))
;                    (display (list 'reset-history historysize))(newline)
                    (if (<= offset (.getCaretPosition both-ta))
                       (.setCaretPosition both-ta (+ (.length str) (.getCaretPosition both-ta))))
                   )))
           (begin
	      (.removeNoSend (.getDocument both-ta) offset len 
                 (lambda R
                    (set! historysize (hist))
                    (let ((x (.getCaretPosition both-ta)))
                        (cond ((< (+ offset len) x)
                            (.setCaretPosition both-ta (- x len)))
                            ((<= offset x) 
                             (.setCaretPosition both-ta offset))
                            (else 'nochange))))))
            )))



     (both-ta
           (apply swingcomp (cons (javax.swing.JTextArea. 
               (grouptool.tatool.PlainDocumentWrapper. insertaction deleteaction)) 
               args)))

     (make-highlighter-both (lambda (the-textarea)
        (let ((highlighter (.getHighlighter the-textarea))
              (greenhighlights (javax.swing.text.DefaultHighlighter$DefaultHighlightPainter. green))
              (bluehighlights (javax.swing.text.DefaultHighlighter$DefaultHighlightPainter. cyan))
              (redhighlights (javax.swing.text.DefaultHighlighter$DefaultHighlightPainter. red))
              (tail 10)
             )
          (lambda (greenoffset blueoffset)
              (.removeAllHighlights highlighter)   
    
              ;; add red highlights if needed
              (if (< (- (max greenoffset blueoffset) (min greenoffset blueoffset)) tail)
          	  (.addHighlight highlighter 
          	     (- (max greenoffset blueoffset) tail)
          	     (min greenoffset blueoffset)
          	     redhighlights))
          
              ;; add green highlights
              (.addHighlight highlighter 
                  	(max 0 (- greenoffset tail))
                         greenoffset
              	         greenhighlights)

              ;; add blue highlights
              (.addHighlight highlighter 
          	         (max 0 (- blueoffset tail))
          	         blueoffset
          	         bluehighlights)
           ))))

     (highlight (make-highlighter-both both-ta)))
       


   (.setRows both-ta rows)
   (.setColumns both-ta cols)
   
   ;;;;;;;;;;;;;;;;
   ;;;listeners
   ;;;;;;;;;;;;;;;;
   

   (G 'add-listener (list name 'get-contents 'both-textarea) (lambda(cmd) 
       (java.lang.Thread.sleep 100L)
       (G 'send (list name 'contents 'both-textarea) (.getText both-ta))))

   (set! greenoffset 0)
   (set! blueoffset 0)

   (let
       ((initialized #t))
       
     (G 'add-listener (list name 'modify-highlights 'both-textarea) 
      (lambda(cmd insoffset inscolor len)
	(begin 
	      (if (equal? inscolor 'green)
		  (begin
		    (set! greenoffset (+ insoffset len))
		    (if (> blueoffset insoffset)
			  (set! blueoffset (+ blueoffset len))))
		  (begin
		    (set! blueoffset (+ insoffset len))
		    (if (> greenoffset insoffset)
			(set! greenoffset (+ greenoffset len)))))
	      (highlight greenoffset blueoffset)

)))


    ;; the green user listens to all insert/delete operations
    ;; from herself and others and merges them into a coherent stream which is then
    ;; sent out on the "logger" key
    ;; the non-green users listen to the logger key and perform these operations
    ;; keeping track of how many they have received, the so-called historysize. 
    ;; Any new requests they send are sent relative to the history size, and the
    ;; green user uses this info to obtain a sensible merge.
    ;; This has not been extensively field tested yet...
   (if (equal? color 'green)
    (begin
     (G 'add-listener (list name 'insert 'both-textarea) (lambda(cmd offset inscolor str histpoint)
       (let ((trueoffset (first (compute-true-operation 'insert histpoint offset (.length str)))))
        (.add history (list 'insert trueoffset (.length str) str inscolor))
        ;; the green user sends logging messages
        (G 'send (list name 'logger 'both-textarea)  'insert trueoffset str (.length str) inscolor (.size history))

        (do-op 'insert trueoffset str (.length str) (lambda() (.size history)))
;        (display (list 'do-op 'insert trueoffset str (.length str) (lambda() (.size history))))(newline)

        (G 'send (list name 'modify-highlights 'both-textarea)  (+ trueoffset 1) inscolor (.length str))
       )))

     (G 'add-listener (list name 'delete 'both-textarea) (lambda(cmd offset inscolor len str histpoint)
       (let* ((trueoperation (compute-true-operation 'delete histpoint offset len))
              (trueoffset (first trueoperation))
              (truelen (second trueoperation)))
         (if (< truelen 1)
          (begin #t ;(display (list "omitted delete" offset len str trueoperation))(newline)
          )
          (begin
            (.add history (list 'delete trueoffset truelen str inscolor))
            ;; the green user sends logging messages
            (G 'send (list name 'logger 'both-textarea)  'delete trueoffset "" truelen inscolor (.size history))
            (do-op 'delete trueoffset "" truelen (lambda() (.size history)))
;            (display (list 'do-op 'delete trueoffset "" truelen (lambda() (.size history))))(newline)
            (G 'send (list name 'modify-highlights 'both-textarea)  (+ trueoffset 1) inscolor (- truelen))
       )))))
   ))


    (G 'add-listener (list name 'logger 'both-textarea) (lambda (cmd action offset str len inscolor hist)

     (if (not (equal? color 'green))
       (begin
          (do-op action offset str len (lambda() hist))
;          (display (list 'do-op-client action offset str len inscolor (lambda() hist))) (newline)
        ))

     ))



     (G 'add-listener (list name 'contents 'both-textarea) (lambda(cmd x)
         (.setText both-ta x)
         (set! initialized #t)))

   both-ta))))






