;; Networking.scm
;; Tim Hickey
;; 10/08/2001
;;
;; This file defines constructors for three Scheme objects
;; which form the low level communication layer for
;; our scheme groupware system.
;;
;;  (define S  (make-servent PORTNUMBER))
;;  (define GS (make-group-server PORTNUMBER))
;;  (define GC (make-group-client NAME GROUP MACHINE PORTNUMBER))
;;
;; These are all implements as scheme procedures which maintain some state
;; and which can be called in an object-oriented manner. For example,
;;    (S 'send '(127.0.0.1 54321) 'test "hi")
;; shows how to have the servent S send a message to the given host/port.
;;
;; This file also provides support for event logging which is
;; automatically enabled by default and which writes to a file
;; in the logs directory by default. The logging procedure is
;;
;;  (logger a b ... z)
;; This procedure appends a line of the form
;;    (TIME a b ... z)
;; to the logfile. It is used by the three networking objects described
;; above and one should avoid using it explicitly until we build a better
;; user level interface for it.
;;
;; A Servent, S, opens a serversocket and then processes messages.
;; Typical Messages have the form:
;;      message\nSchemeTerm\n
;; The servent reads the SchemeTerm and closes the connection.
;;
;; Each of these objects is a closure which can be sent messages:
;; (S 'add-listener KEY (lambda R ...))
;; (S 'logout)
;; (S 'send (list IPADDFRESS PORT) KEY A1 A2 ... AN)
;; (S 'get 'address)
;; (S 'get 'port)
;; (S 'get 'ip)
;; (S 'get 'listeners)
;; (S 'get 'serversocket)
;; (S 'set 'address (list MACHINE PORT))
;;
;; A group-server, GS, opens a serversocket
;; which clients connect to when they want to join the group
;; or query the group. 
;; When a member sends a message to the server it is
;; rebroadcast to all members of the group (self-included)
;;
;; If you access the server via telnet, the socket protocol is
;; join\nUSER\nGROUP      -- enroll the USER in the GROUP with this socket
;; quit\n                  -- stop the server
;; groups\n               -- get list of groups
;; users\n                -- get list of users
;; users-in-group\nGROUP  -- get list of users in the group
;;
;; The group-server object accepts several messages
;; (GS 'get 'thread)  -- gets current socket-handler-thread
;; (GS 'get 'peers)
;; (GS 'get 'count)
;; (GS 'get ss)
;; (GS 'quit)
;;
;;
;; The group-client provides a high-level interface to the
;; group-server protocol. It accepts the following messages
;;
;;
;; (define GC (make-group-client 'mary 'cs2a "frege.cs-i.brandeis.edu" 54321))
;; (GC 'add-listener 'abc (lambda R (display (list "got abc msg" R))))
;; (GC 'send KEY A B C ...)  --> send a message to everyone in the group (self-included)
;; (GC 'get 'address)
;; (GC 'get 'socket-info)  --> (list socket in out)
;; (GC 'get 'listeners)
;; (GC 'get 'thread)     --> socket-handler thread
;; (GC 'logout) 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define logging #t)
(define log-file 
 (string-append 
    "logs/networking-"
    (java.lang.System.currentTimeMillis)
   ".log"))

(define log-stream 
  (if logging (java.io.PrintWriter. (java.io.FileWriter. log-file))
      #null))

(define (close-log) (.close log-stream))

(define logger (lambda R
  (if logging
    (begin
       (.println log-stream (toStringQ 
          (cons (java.lang.System.currentTimeMillis) R)))
       (.flush log-stream)))))

(define debug #f)
(define printdebug (lambda R 
  (if debug (begin (display "DEBUG:>> ") (display R)(newline)))))

(define (toStringQ S)  (jscheme.REPL.printToString S #t))
(define string->exprlist jscheme.REPL.parseScheme)

(define (make-servent port)
  ;; This is a very simple serverloop
  ;; it accepts sockets  and spawns a thread to handle the socket.
  ;; The thread immediately reads a string on the socket to get the message type
  ;; then it reads the next line and converts it to a scheme term
  ;; handles the term and closes the connection.

  (define ss (java.net.ServerSocket. port))

  (define (start-server)
    (let serverloop ((s (.accept ss)))
        (.start (java.lang.Thread. (lambda() (handle-socket s))))
        (serverloop (.accept ss))))

  (define listeners ())
  (define groups-I-own ())
  (define groups-I-have-joined ())

  (define (handle-socket s)
;      (define z (Math.round (* 1000 (java.lang.Math.random))))
;      (display (list "in handle socket" s z))(newline)
         (tryCatch
          (begin 
               (printdebug (tryCatch (list (.getSoLinger s) (.getSoTimeout s) (.getTcpNoDelay s)) (lambda(e)
                    (list "error in printdebug" s e))))
           (let* (
                  (in (java.io.BufferedReader.
                       (java.io.InputStreamReader. 
                         (.getInputStream s))))
                  (out (java.io.PrintStream.
                        (.getOutputStream s)))
                  (message-type (.readLine in)))
;             (display (list "in handle socket 2"  z message-type in out))(newline)
             (printdebug message-type)
             (case message-type
              (("message")
               (let ((theData 
                       (first 
                         (string->exprlist 
                            (tryCatch 
                               (.readLine in) 
                               (lambda(e) (display (list "I/O error" e))(.close s)(throw e)))))))
                   (printdebug (string-append 'just-read theData "\n\n\n"))
                   (logger 'servent 'in theport theData)
                   (apply-super-listeners (first theData) (rest theData) super-listeners)
                   (apply-listeners (first theData) (rest theData) listeners)
                  ))

              ;; there is room here for other message types
              (("ping")
                (.println out (.getHostAddress (.getInetAddress s))))

              (else (warning (list 'unknown-message-type message-type))))
             (.close in)(.close out) (.close s)
;           (display (list "leaving handle socket" z))(newline)
          ))
          (lambda(e) (.close s) (warning (list "exception in handle socket " s e)) )))


  (define (apply-listeners Key Val Listeners)
    (if (not(null? listeners)) (printdebug (list "applying listeners " Listeners " to " Key " and " Val )))
    (for-each 
         (lambda(x) 
            (if (equal? (first x) Key) (apply (second x) (cons Key Val))))
         Listeners))


   (define super-listeners ())

   (define (apply-super-listeners Key Val Listeners)
        (if (not (null? listeners)) (printdebug (list "applying group super listeners " Listeners " to " Key " and " Val )))
        (for-each 
             (lambda(x) 
                (apply (first x) (cons Key Val)))
             Listeners))
    


  (define N 
    (let ((HOSTADDRESS (tryCatch (.getHostAddress (java.net.InetAddress.getLocalHost)) (lambda(e) "localhost"))))
      (lambda R
        (case (first R)

          ((add-listener) 
           (set! listeners (cons (rest R) listeners)))

          ((add-super-listener) 
           (set! super-listeners (cons (rest R) super-listeners)))

          ((logout) 
           (.close ss))

          ((send) 
           (logger 'servent 'out theport R)
           (if (equal? (second R) 'all)
               (broadcast-message (rest R))
               (send-to-user (second R) (rest (rest R)))))
          ((get)
           (case (second R)
            ((address) (list HOSTADDRESS (.getLocalPort ss)))
            ((port) (.getLocalPort ss))
            ((ip) (.getHostAddress (.getInetAddress ss)))
            ((listeners) listeners)
            ((socket-server) ss)
            (else (warning (list "ERROR: unknown variable " (second R))))))

          ((set)
           (case (second R)
             ((host) (set! HOSTADDRESS (third R)))
             (else (warning (list "unknown command" R)))))

          (else
            (warning (list "unknown command" R)))))))



  (define (broadcast-message R)
    (warning "this has not yet been implemented"))

  (define (send-to-user User Rest)
    (tryCatch
     (let ((s (java.net.Socket. (first User) (second User))))
        (let ((in (java.io.BufferedReader.
                     (java.io.InputStreamReader. 
                         (.getInputStream s))))
              (out (java.io.PrintStream.
                      (.getOutputStream s))))
         (.println out "message") 
         (.println out (toStringQ Rest)))
       (.close s))
    (lambda(e) 
       (warning (list "error during send-to-user" User Rest e)))))

  (define (warning X) 
    (display X)(newline))

  (define theport (.getLocalPort ss))
    
  (.start (java.lang.Thread. (lambda() (start-server))))
  N
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; make-group-server
;;
;;
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



(define (make-group-server port)
    
    (define (create-peer group in out s username )
         (list group in out s username (java.lang.System.currentTimeMillis))) 
    
    (define (update-peer-time peer)
      (set-car! (list-tail peer 5) (java.lang.System.currentTimeMillis)))
    (define (get-peer-groups peers) (rem-dups (map first peers) ()))
    (define (get-peer-users peers) (rem-dups (map (lambda(x) (list-ref x 4)) peers) ()))
    (define (get-peer-users-in-group peers group)
      (filter (lambda(x) (if (equal? (first x) group) (list-ref x 4) #f)) peers))
    
    (define (filter F L)
      (if (null? L) ()
         (let ((x (F (first L))))
            (if x (cons x (filter F (rest L))) (filter F (rest L))))))
    
    (define (remove-peer username group P)
       (cond 
          ((null? P) ())
          ((and (equal? username (list-ref (first P) 4)) (equal? group (first (first P))))
           (remove-peer username group (rest P)))
          (else (cons (first P) (remove-peer username group (rest P))))))
    
    (define (broadcast-to-peers peers message)
      (for-each (lambda(x) (.println (third x) (U.stringify message))) peers))  
    
    (define (broadcast-to-group peers group theData)
;     (logger 'group 'server group peers theData) 
     (map (lambda(x) 
            (if (equal? (first x) group)
                (begin
;                   (logger 'group 'server-send group (cons (first x) (rest (rest (rest (rest x))))) theData) 
                   (.println (third x) theData)
                   (.flush (third x)))))
          peers)
    )
    
    ;; This is a very simple serverloop
    ;; it accepts sockets and immediately reads a string on the socket to get the User ID
    ;; then it starts a forwarding thread which forward everything that comes into the socket
    ;; to all the current peers (including the new socket itself).
    (define peers ())
    (define count 0)
    (define ss (java.net.ServerSocket. port))

    (define (start-server port)
        (let serverloop ()
           (let ((s (.accept ss)))
            (.start (java.lang.Thread. (lambda()
					 (tryCatch
              (begin ;(display (list (.getSoLinger s) (.getSoTimeout s) (.getTcpNoDelay s))) (newline) 
               (let* (
                      (in (java.io.BufferedReader.
                           (java.io.InputStreamReader. 
                             (.getInputStream s))))
                      (out (java.io.PrintStream.
                            (.getOutputStream s)))
                      (action (.readLine in)))
                 ;(display action)(newline)
                 (case action
    
                  (("join")
                    (let* (
                      (username (.readLine in))
                      (group (.readLine in))
                      (peer (create-peer group in out s username))
                      (forwarding-thread 
                            ;; this waits for messages from the user and forwards them to peers
                       (java.lang.Thread.
                          (lambda() 
                           (let loop ((c 0))
                             (let ((theData (tryCatch (.readLine in) (lambda(e) #null))))
;                               (display (list 'zzzz username group theData))(newline)
                               (cond 
                                 ((equal? theData #null)
                                     ;; remove this element from the list of peers
                                     (synchronize 'p
                                       (lambda(p) (set! peers (remove-peer username group peers)))))
                                 ((equal? theData "")
                                  ;(display (list "ack" username)) (newline)
                                  (loop (+ c 1))) 
                                 (else
                                     (set! count (+ 1 count))
                                     ;;; broadcast the data to the peers in the group
                                     (broadcast-to-group peers group theData)
                                     (loop (+ c 1)))))))
                           username)))
;                        (display (list "BEFORE: The peers are: " peers))(newline)(newline)

                        (synchronize 'p (lambda(p) 
                           ;;; add a new peer
                           (set! peers (cons peer peers))))
;                        (display (list "AFTER: The peers are: " peers))(newline)(newline)
                        (.start forwarding-thread)))
    
                  (("groups")
                    ;;(display "handling groups")(newline)
                    ;; print group names 
                    (.println out (U.stringify (get-peer-groups peers)))
                    (.println out "\n")
                    (.flush out)   
                    (.close s)
                   )
    
                  (("users")
                    ;;(display "handling groups")(newline)
                    ;; print group names 
                    (.println out (U.stringify (get-peer-users peers)))
                    (.println out "\n")
                    (.flush out)   
                    (.close s)
                   )
                  (("users-in-group")
                    ;;(display "handling groups")(newline)
                    ;; print group names 
                    (.println out (U.stringify (get-peer-users-in-group peers (.readLine in))))
                    (.println out "\n")
                    (.flush out)   
                    (.close s)
                   )
    
                  (("quit")
                    (java.lang.System.exit 0))
    
                  (else
                    (.println out "(ERROR unknown action:)")
                    (.println out action)
                    (.flush out)
                    (.close s)))))
                (lambda(e) (display (list "ERROR: " e)) (newline)))
               ))))
             ;;(display "starting new loop") (newline)
             (serverloop)  ))
    
           
    ;; merge list L with list S with no duplicates being added to S
    (define (rem-dups L S)
      (if (null? L) S
        (let ((x (first L)))
          (if (member x S)
              (rem-dups (rest L) S)
              (rem-dups (rest L) (cons x S))))))
    
    
    (define server-thread 
       (java.lang.Thread. (lambda() (start-server port))))

    (.start server-thread)

  (lambda R
     (case (first R)
       ((get)
           (case (second R)
              ((thread) t)
              ((peers) peers)
              ((count) count)
              ((ss) ss)
              ((port) (.getLocalPort ss))
              (else (list "don't understand" R))))
       ((quit)
          (.close ss)
          (.stop server-thread))))
)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; make-group-client
;;  super-listeners listen to all messages
;; USE:
;;   (define G (make-group-client 'tim 'mygroup "localhost" 1118))
;;   (G 'add-listener 'abc (lambda R (display (list "abc just heard " R))(newline)))
;;   (G 'add-super-listener  (lambda R (display (list "super heard:" R))(newline)))
;;   (G 'send 'abc 1 2 3 4)
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-group-client name group host port)
    ;; This is a very simple serverloop
    ;; it accepts sockets  and spawns a thread to handle the socket.
    ;; The thread immediately reads a string on the socket to get the message type
    ;; then it reads the next line and converts it to a scheme term
    ;; handles the term and closes the connection.
    
    
      (define (toStringQ S)  (jscheme.REPL.printToString S #t))
      (define string->exprlist jscheme.REPL.parseScheme)
    
      (define s (java.net.Socket. host port))

      (define in (java.io.BufferedReader.
                           (java.io.InputStreamReader. 
                             (.getInputStream s))))
      (define out (java.io.PrintStream.
                            (.getOutputStream s)))
    
    
      (define (group-thread-handler)
         (let loop ()
              (let ((x (tryCatch (first (string->exprlist (.readLine in))) (lambda(e) ()))))
                (if (pair? x)
                   (begin
                      (tryCatch 
                        (begin
                            (logger 'group 'in (list name host port theport) x) 
                            (apply-super-listeners (first x) (rest x) super-listeners)
                            (apply-listeners (first x) (rest x) listeners))
                        (lambda(e) ;; if there is an error, 
                            (warning (list "group-thread-handler error: " e))))
                      (loop))
                   (.close s)))))
;""))
;            (loop)))
    
      (define listeners ())

      (define (apply-listeners Key Val Listeners)
        (if (not(null? Listeners)) (printdebug (list "applying group listeners " Listeners " to " Key " and " Val )))
        (for-each 
             (lambda(x) 
                (if (equal? (first x) Key) (apply (second x) (cons Key Val))))
             Listeners))
    
      (define super-listeners ())

      (define (apply-super-listeners Key Val Listeners)
        (if (not (null? listeners)) (printdebug (list "applying group super listeners " Listeners " to " Key " and " Val )))
        (for-each 
             (lambda(x) 
                (apply (first x) (cons Key Val)))
             Listeners))
    

      (define handler-thread (java.lang.Thread. group-thread-handler))

      (define G 
          (lambda R
            (case (first R)
    
              ((add-listener) 
               (set! listeners (cons (rest R) listeners)))

              ((add-super-listener) 
               (set! super-listeners (cons (rest R) super-listeners)))
    
              ((logout) 
               (.close s))
    
              ((send) 
               (logger 'group 'out theport R) 
               (.println out (toStringQ (rest R))))
    
              ((get)
               (case (second R)
                ((address) (list host port))
                ((socket-info) (list s in out))
                ((listeners) listeners)
                ((thread) handler-thread)
                (else (warning (list "ERROR: unknown variable " (second R))))))
    
              (else
                (warning (list "unknown command" R))))))
    
    
    
    
      (define (send-to-group User Rest)
        (tryCatch
         (let ((s (java.net.Socket. (first User) (second User))))
            (let ((in (java.io.BufferedReader.
                         (java.io.InputStreamReader. 
                             (.getInputStream s))))
                  (out (java.io.PrintStream.
                          (.getOutputStream s))))
             (.println out "message") 
             (.println out (toStringQ Rest)))
           (.close s))
        (lambda(e) 
           (warning (list "error during send-to-user" User Rest e)))))
    
      (define (warning X) 
        (display X)(newline))
        
  (define theport (.getPort s))

  (display (list "group-client joining group" name group)) (newline)
  (.start handler-thread)
  (.println out "join")
  (.println out name)
  (.println out group)
  G
)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; (send-line host port msg)
;;
;; This is a telnet-like interface that lets you send some lines
;; to a host/port and read one line, after which the socket is closed.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(define (send-line host port msg)
   (let* (
          (s (java.net.Socket. host port))
          (theport (.getLocalPort s))
          (in (java.io.BufferedReader.
               (java.io.InputStreamReader. 
                (.getInputStream s))))
          (out (java.io.PrintStream.
                (.getOutputStream s))))
     (logger 'send-line host theport msg)
     (.println out msg)
     (let ((x (.readLine in)))
       (.close s)
       x)))


