;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; registrar.scm
;; author: Tim Hickey
;; date: 1/27/2002
;;   rewrote the code so that it uses an OO model
;;   also rewrote the comments
;;
;;  % java -cp jscheme.jar jscheme.REPL tat/Networking.scm tat/Registrar.scm '(make-registrar 23232 "thanks")'
;;
;; this creates a servent and uses it to manage a list of active users
;; the users can login/logout and request information about
;; other users (e.g. current host/port of one or all users)
;; the server also implements a leasing policy 
;; by requiring users to respond to periodic
;; messages. It removes them if they do not respond in a
;; timely manner.
;;
;;
;; (make-registrar PORT PASSWORD) creates a registrar object R with the following methods:
;; (R 'get 'servent) -- returns the underlying servent
;; (R 'get 'activeusers) -- returns the list of activeusers
;;
;; Summary of the registrar protocol for a servent N
;; connecting to the registrar at address REG 
;;
;; logging in:
;;  (N 'send REG 'login MYNAME MYPASSWORD (MYHOST MYPORT))
;; listeners
;;  (N 'add-listener 'login-successful (lambda(cmd) ...))
;;  (N 'add-listener 'add-user         (lambda(cmd username) ...))
;;  (N 'add-listener 'login-failed     (lambda(cmd) ...))
;;
;; logging out:to server
;;  (N 'send REG 'logout MYNAME MYPASS)
;; listeners:
;;  (N 'add-listener 'logout-successful (lambda(cmd)...))
;;  (N 'add-listenerr'drop-user         (lambda(cmd username) ...))
;;
;; broadcasting to all users:
;;  (N 'send REG 'broadcast MYNAME MYPASS A1 A2 ... AN)
;; listeners:
;;  (N 'add-listener 'broadcast-A1 (lambda (cmd . Args) ... ))
;;
;; lookup-service
;;  (N 'send REG 'lookup SOMEONE MYNAME MYPASS)
;; listeners:
;;  (N 'add-listener 'user (lambda(cmd THEIRNAME THEIRADDRESS) ... ))
;;
;; lookup all users
;;  (N 'send REG 'lookupall MYNAME MYPASS)
;; listeners:
;;  (N 'add-listener 'userlist (lambda(cmd L) ... ))
;;    where L has the form: ((N1 (H1 P1)) (N2 (H2 P2)) ... ))
;;    for users Ni with hosts Hi and ports Pi
;;
;; required listener:
;;  (N 'add-listener 'areyouthere 
;;     (lambda(cmd)
;;       (N 'send REG 'iamhere MYNAME)))
;;
;;
;; Note: we assume here that there are no malicious users
;; In a second pass we will try to add security to prevent
;; malicious users from wreaking havoc.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (go) 
  (newline)(newline)
  (display "starting registrar with password 'thanks' on port 23232")
  (newline)(newline)
  (make-registrar 23232 "thanks"))

(define (make-registrar port authorizor)
    (define N (make-servent port))

    (define authorize 
      (if (string? authorizor)
          (lambda (name password) (equal? password authorizor))
          authorizor))

    (define (now) 
        (java.lang.System.currentTimeMillis))
    
    (define activeusers ())  ; list of   (name (IP port) lastcontact)

    (define (lookup x L)
      (if (null? L)
          #null
          (if (equal? (first (first L)) x) 
              (second (first L))
              (lookup x (rest L)))))
    
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ;; leaseholder
    ;; this periodically loops through the list of activeusers
    ;; and sends them messages to make sure they are active.
    ;; If a process misses three of these wake-up calls, it is
    ;; removed from the activeusers list.
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    (define leaseholder 
     (let ((leasetime 20000L))
      (java.lang.Thread. (lambda()
       (let loop ()
         (java.lang.Thread.sleep leasetime)
         (for-each 
           (lambda(x)
             (let ((interval (- (now) (third x))))
    ;             (display (list 'inloop x interval leasetime))(newline)
                 (if (> interval (* 3 leasetime))
                     (remove-active-user (first x))
                     (N 'send (second x) 'are-you-there)))) 
            activeusers)
         (loop))))))
    
    ;; These are helper functions for the leaseholder thread
    (define (update-last-activity who when users)
    ;  (display (list 'ula who when users))(newline)
      (if (null? users) "ERROR"
          (if (equal? (first (first users)) who)
              (set-car! (rest (rest (first users))) (now))
              (update-last-activity who when (rest users)))))
    
    (define (remove-active-user x)
      (synchronize 'a (lambda(a) 
         (set! activeusers (remove-user x activeusers))
         (broadcast (list 'drop-user activeusers x))
      )))

    (define (add-active-user x)
      (synchronize 'a (lambda(a) 
         (set! activeusers (cons x activeusers))
         (broadcast (list 'add-user activeusers x))
      )))

    (define (remove-user x L)
      (if (null? L) L
          (if (equal? (first (first L))  x)
              (rest L)
              (cons (first L) (remove-user x (rest L))))))
      
    (define broadcast (lambda (R)
        (for-each 
           (lambda (x) 
              (apply N `(send ,(second x) ,@R)))
           activeusers)))


    (N 'add-super-listener (lambda R (display (list 'sys R))(newline)))

    (N 'add-listener 'login (lambda(cmd name password address)
       (if (not (authorize name password))
           (begin
             (display (list 'login-denied name password))(newline)
             (N 'send address 'login-denied))
           (begin
             (N 'send address 'login-successful)
             (add-active-user (list name address (now) password))
           ))))
    
    (N 'add-listener 'i-am-here (lambda (cmd who) 
       (update-last-activity who (now) activeusers)))
    
    (N 'add-listener 'logout (lambda (cmd name password) 
       (if (not(authorize name password))
           (begin 
             (display (list "autherror" cmd name password))(newline))
           (begin
             (N 'send (lookup name activeusers) 'logout-successful)
             (remove-active-user name)
             ))))
    
    (N 'add-listener 'lookup (lambda (cmd who name password)
       (if (not (authorize name password))
          (begin (display (list "autherror" cmd who name password))(newline))
          (begin
            (N 'send (lookup name activeusers) 
               'user who (lookup who activeusers))))))
    
    (N 'add-listener 'broadcast (lambda (cmd name password a1 . args)
       (if (not (authorize name password))
          (begin (display (list "autherror" cmd who name password))(newline))
          (broadcast (cons (string-append "broadcast-" a1) args)))))
    
    (N 'add-listener 'lookup-all (lambda (cmd name password) 
       (if (not (authorize name password))
          (begin (display (list "autherror" cmd name password))(newline))
          (begin
            (N 'send (lookup name activeusers)
               'userlist (map first activeusers))))))

    (.start leaseholder)

  ;; here is where we construct the registrar object itself...
 (lambda (cmd . args)
    (case cmd
       ((get)
         (case (first args)
            ((servent) N)
            ((activeusers) activeusers)
            ((leaseholder) leaseholder)
            (else (list 'error "unknown args for get:" args))))
       (else (list 'error "unknown command:" cmd args))))
)


