; VCR.scm
; this creates a VCR object that creates a standard VCR control panel
; and processes the eventlist by sending commands off to the appropriate
; event handlers
;
; The VCR reads an event stream and uses the event stream to control the components.
; The events have the form
;  (Time (C1 C2 ... Cn) E D1 ... Dm)
; where 
; Time is the time of the event (in milliseconds since 1/1/1970)
; E is the type of the event, and
; C1 C2 ... Cn is the sequence of components in which this event occurs
; (C1 is the outermost component, which contains C1, etc.)
; and D1 ... Dm are the component-specific data describing the event
; with enough detail to replay it and also to undo it.
; Each component has its own hander which is able to process events of the form
; (direction  (C1 C2 ... Cn) E D1 ... Dm)
; where direction is 'forward or 'backward
;
; Thus to construct and run a vcr for a component C you need to
; 1) construct the component C in 'vcr mode
;     (define C (make-COMPONENT))
; 2) Get the vcrhandler H for the component
;     (define H (C 'get 'vcrhandler))
; 3) Create the vcr object and pass it the vcrhandler
;     (define V (make-vcr 'handler H))



(define make-vcr (lambda R 
  ; global variables
  (define events #())   ;; this is the array of events that will be analyzed
  (define numevents 0)  ;; this is the size of the event array
  (define handler (lambda R #t)) ;; this is the default event handler
  (define t (maketagger))
  (define start-time (.getTime (Date.)))
  (define end-time start-time)
  (define pos 0)
  (define the-time (.getTime (Date.)))
  (define update (lambda ()
           (writeexpr (t "time") (Date. (.longValue the-time)))
           (writeexpr (t "status") vcrmode)
           (writeexpr (t "elapsedtime") (* 0.001 (- the-time start-time)))
           (writeexpr (t "delay") delay)
           (writeexpr (t "eventnum") pos)))
  (define vcr-error (lambda R (writeexpr  (t "status") {[R]})))

  (define speedslider
    (let ((this (javax.swing.JSlider. javax.swing.SwingConstants.HORIZONTAL$ -100 100 10 )))
       (.addChangeListener this (jsint.Listener. (lambda(e)
             (if (not (.getValueIsAdjusting this))
                 (let ((v (.getValue this)))
		   (if (< v 0)
                       (run 'rewind (/  v -10.0))
                       (run 'play (/  v 10.0))))))))
       this))


  (define (num->string num decimals)
    {[(.intValue (floor num))].[(.intValue (round (* (Math.pow 10 decimals) (- num (.intValue num)))))]})

  (define controller
    (window "vcr"
     (border 
       (north 
        (col
         (row (label "LogFile:") 
               (t "LogFile" 
                 (textfield "" 30 (action (lambda(e)
                    (initialize-VCR (readstring (t "LogFile")))))  ))
              (button "browse" (action (lambda(e)
                  (.show FileDialogLoad)
                  (writeexpr (t "LogFile")
                    (.getAbsolutePath
                       (java.io.File. 
                           (.getDirectory FileDialogLoad) 
                           (.getFile FileDialogLoad))))
                  (initialize-VCR (readstring (t "LogFile")))
                  (writeexpr (t "status") "ready")
                  (writeexpr (t "eventnum") 0)
                  ))))
         (col
          (row (label "time")  
           (t "time" (label "-----------------------------------------")))
           (row (label "event")  
           (t "event" (label "-----------------------------------------")))
           (row (label "status") 
           (t "status" (label "" (TimesRomanItalic 12)))))))

       (center 
        (row 'none 'west
;         (t "evt" (choice "none" "chat"))
         (table 2 5
           (label "event number") 
           (label "lastevent")
           (label "elapsed time")
           (label "total time")
           (label "time to next event")
           (t "eventnum" (textfield "" 5))
           (t "lastevent" (label "-----"))
           (t "elapsedtime" (label "-----"))
           (t "totaltime" (label "-----"))
           (t "delay" (label "-----")))))
         

       (south
        (col 'none 'center
         (border (west (t "vcrspeed" (textfield "0.0" 10))) (center speedslider))
         (grid 1 7
          (button "|<" (action (lambda(e) (run 'rewind 10000000.0))))
          (button "<<" (action (lambda(e) (run 'rewind 10.0))))
          (button "<-" (action (lambda(e) (run 'rewind 1.0))))
          (button "<|" (action (lambda(e) (run 'step-backward 1.0))))
          (button "||" (action (lambda(e) (set! vcrmode 'stop))))
          (button "|>" (action (lambda(e) (run 'step-forward 1.0))))
          (button "->" (action (lambda(e) (run 'play 1.0))))
          (button ">>" (action (lambda(e) (run 'play 10.0))))
          (button ">|" (action (lambda(e) (run 'play   10000000.0))))))))))
  
  
  (define vcrmode 'step)
  (define vcrspeed 1.0)
  (define (run mode speed)
      (define (flip x) (case x ((play) 'rewind) ((rewind) 'play) (else x)))
      (set! delay 0)
      (set! vcrspeed speed) ;(* speed (abs (.getValue speedslider))))
      (set! vcrmode mode) ; (if (>= (.getValue speedslider) 0)			mode			(flip mode))
      (writeexpr (t "vcrspeed") (num->string vcrspeed 2))
   )

  (define (stepbackward)
       (if (< pos 1) (vcr-error "at start of tape")
             (begin
               (set! the-time (first (vector-ref events pos)))
               (set! pos (- pos 1))
               (writeexpr (t "event") (vector-ref events pos))
               (update)
               (player pos 'backward) 
             )))


  (define (stepforward)
       (if (>= pos (- numevents 1)) (vcr-error "at end of tape")
             (begin 
               (player pos 'forward)
               (tryCatch (set! the-time (first (vector-ref events pos))) (lambda(e) #t)) ;; catch array out of bounds
               (writeexpr (t "event") (vector-ref events pos))
               (set! pos (+ pos 1))
               (update)  )))

  ;; there are used to simulate long waits and still allow the process
  ;; to be interrupted by changing the value of vcrmode to 'stop

  (define delay 0)
  (define pollinterval 250) ; milliseconds

  ;; this is the VCR player thread, it is always on
  (define play-vcr 
    (lambda ()
     ;; if in a long delay then just wait for 0.25 secs and decrement delay
     (if (and (> delay 0) (not (equal? vcrmode 'stop)))
         (begin
           (Thread.sleep (.longValue (min  pollinterval  delay)))
           (set! delay (max 0 (- delay  pollinterval )))
           (update)
           (play-vcr)))

     ;; otherwise, move forward or backward
     (case vcrmode
        ((play) (stepforward))
        ((rewind) (stepbackward))
        ((step-forward) (stepforward) (set! vcrmode 'stop))
        ((step-backward) (stepbackward) (set! vcrmode 'stop)))


     ;; if not at the end, and moving forward or backward
     ;; calculate the length of the next move and go
     (if (and (< pos (- numevents 1))
              (> pos 0) 
              (or (equal? vcrmode 'rewind) (equal? vcrmode 'play)))
         (let ((wait
                   (/ (Math.abs (- (first (vector-ref events pos)) the-time))
                      vcrspeed)))
            (set! delay (max 0 (- wait  pollinterval )))
            (Thread.sleep (.longValue (min  pollinterval  wait))))
         (Thread.sleep  (.longValue pollinterval)))
      (play-vcr)))


  (define FileDialogLoad 
      (tryCatch (java.awt.FileDialog. controller "VCR Log File Loader" java.awt.FileDialog.LOAD$)
                  (lambda(e) #null)))


  ;; This reads the scheme expressions from the filename, stores them in an array, and initializes
  ;; the VCR for this data
  (define (initialize-VCR filename)
    (tryCatch
      (let* ((F (open-input-file filename))
             (these-events 
                (let loop ((L ())) 
                  (let ((x (read F)))
;                    (display (list "reading" x))(newline)
                    (if (eof-object? x) 
                        (reverse L)
                        ;(loop (cons (string->expr x) L)))))))
                        (loop (cons x L)))))))
;         (display (list "just read " these-events " from " F ))(newline)
         (set! events (list->array Object.class these-events))
         (set! numevents (Array.getLength events))
         (set! start-time (first (first these-events)))
         (set! end-time (first (vector-ref events (+ -1 numevents))))
         (writeexpr (t "totaltime") (num->string (* 0.001 (- end-time start-time)) 3))
         (writeexpr (t "lastevent") (+ -1 (Array.getLength events))))
      (lambda(e) 
          (vcr-error {\n\nError during read: [ e ]\n\n})
          'error)))

  ;; here we assume that the log file has been stored in an array
  ;; events, where each event is a list whose first element is the event time in milliseconds
  ;; and we assume we have a general event handler
  ;;  (handler direction event)
  ;; which will handle the event

  (define (player Position Direction)
    (case Direction
      ((forward)
         (if (equal? Position numevents)
             (vcr-error "at end of tape, can't move forward")
             (begin
;               (display (list 'player Direction Position (rest (vector-ref events Position))))(newline)
               (handler () 'set 'direction 'forward) ;; () for all contexts
               (apply handler  ;; event has form (time context eventtype data data ...)
                  (rest        ;; strip off time filed
                    (vector-ref events Position))))))
      ((backward)
         (if (< Position 0)
             (vcr-error "at beginning of tape, can't move backward")
             (begin
;               (display (list 'player Direction Position (vector-ref events Position)))(newline)
               (handler () 'set 'direction 'backward) ;; () for all contexts
               (apply handler (rest (vector-ref events Position))))))))

  (define msghandler
        (lambda (cmd . args)
           (case cmd
             ((get)
	       (case (first args)
		 ((controller) controller)
                 (else #t)))

	     ((setall)     ;; this allows multiple sets: (P 'setall 'name "Tim" 'color blue)
              (let loop ((L args)) 
                   (if (null? L) #t
                   (begin 
                       (msghandler 'set (first L) (second L)) 
                       (loop (rest (rest L)))))))

	     ((set)
	       (case (first args)
                  ((events) 
                     (set! events (list->array Object.class (second args)))
                     (set! numevents (Array.getLength events)))
                  ((handler)
                     (set! handler (second args)))
                  (else #t)))

             (else (display (list "ERROR" cmd args)))
	   )))
     (apply msghandler (cons 'setall R)) ;; interpret parameters as setting specifiers
     (.start (Thread. play-vcr))
     msghandler))




